#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2015 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#

try:
    import logging
    import logging.handlers
    from   nlmanager.nllistener import NetlinkManagerWithListener
    import nlmanager.nlpacket
    import os
    import Queue
    import select
    import socket
    import struct
    import threading
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

class ClagNl():

    # Event code constants
    _GET_ALL_ADDRESSES      = 1
    _GET_ALL_LINKS          = 2
    _GET_ALL_NEIGHBORS      = 3
    _GET_ALL_MACS           = 4
    _GET_ALL_ROUTES         = 5
    _SHUTDOWN               = 6
    _GET_ALL                = 7
    _SERVICE_NETLINK_QUEUE  = 'SERVICE_NETLINK_QUEUE'
    _SERVICE_ERROR          = 'SERVICE_ERROR'

    def __init__(self, Daemon, Log, Intf, Parser):
        self.Daemon = Daemon
        self.Log = Log
        self.Intf = Intf
        self.Parser = Parser
        self.FdbSync = None
        self.NeighSync = None

        # Data shared by all the threads
        self.keep_going = True

        self.nlm = None
        self.netlinkThread = threading.Thread(None, self.NetlinkThreadT, "NetlinkThreadT")
        self.netlinkThread.daemon = True
        self.netlinkThread.start()

    def __del__(self):
        # Signal the update thread to wake up and exit
        self.SignalThreads()
        self.JoinThreads()

    def SignalThreads(self):
        self.keep_going = False
        if self.nlm:
            self.nlm.workq.put((ClagNl._SHUTDOWN, None))
            self.nlm.listener.shutdown_event.set()

    def JoinThreads(self):
        if self.nlm:
            self.nlm.listener.join()
        self.netlinkThread.join()

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_NETLINK, logStr)

    def log_verbose(self, logStr):
        self.Log.log_verbose(self.Log._LOG_DEBUG_NETLINK, logStr)

    def NetlinkThreadT(self):
        try:
            self.Log.log("[%d]Init NetlinkThreadT" % (self.Parser.getTid()))
            self.NetlinkThread()
        except:
            self.Daemon.DumpTraceback()

    def SetFdbSync(self, FdbSync):
        self.FdbSync = FdbSync
        if self.nlm is not None:
            self.nlm.FdbSync = FdbSync

    def SetNeighSync(self, NeighSync):
        self.NeighSync = NeighSync
        if self.nlm is not None:
            self.nlm.NeighSync = NeighSync

    def DumpMacs(self):
        if self.nlm is not None:
            self.nlm.workq.put((ClagNl._GET_ALL_MACS, None))

    def DumpNeighs(self):
        if self.nlm is not None:
            if self.Parser.args.permanentMacSync:
                self.nlm.workq.put((ClagNl._GET_ALL_ADDRESSES, None))
            self.nlm.workq.put((ClagNl._GET_ALL_NEIGHBORS, None))

    def mac_int_to_str(self, mac_int):
        """
        Return an integer in MAC string format: xx:xx:xx:xx:xx:xx
        """
        return ':'.join(("%012x" % mac_int)[i:i+2] for i in range(0, 12, 2))

    def service_overflow(self):
        """
        When an overflow is detected, stop the netlink listener, add a stale
        flag on all the ifaces, restart the listener, request a link dump
        which will clear the stale flag and then delete all stale ifaces
        """
        self.log_debug('NetlinkThread: Netlink overflow, trying to resync')
        self.nlm.listener.shutdown_event.set()
        self.nlm.listener.join()

        # flush the current workq - if shutdown is detected just bail
        while True:
            try:
                (event, options) = self.nlm.workq.get_nowait()
            except Queue.Empty:
                break
            if event == ClagNl._SHUTDOWN:
                self.log_debug("NetlinkThread: shutting things down")
                return
        self.Intf.ifInfoLock.acquire()
        for ifDict in self.Intf.ifInfoByIndex.itervalues():
            ifDict['stale'] = True
        self.Intf.ifInfoLock.release()
        # TODO 4.0: Fdb resync
        if self.NeighSync:
            self.NeighSync.neighDBLock.acquire()
            for neigh in self.NeighSync.neighDB:
                self.NeighSync.neighDB[neigh][self.NeighSync.NEIGH_DB_STALE] = True
            self.NeighSync.neighDBLock.release()

        self.nlm.restart_listener()
        self.nlm.listener.supported_messages_add(nlmanager.nlpacket.NLMSG_DONE)
        self.nlm.workq.put((ClagNl._GET_ALL, None))

    def NetlinkThread(self):
        """
        This function operates as an independent thread of the script.
        """

        self.log_debug("Beginning execution of the thread NetlinkThread")

        # These are the groups we are interesting in listening to.
        groups = nlmanager.nlpacket.RTMGRP_LINK | \
                 nlmanager.nlpacket.RTMGRP_NEIGH
        if self.Parser.args.permanentMacSync:
            groups |= nlmanager.nlpacket.RTMGRP_IPV4_IFADDR | \
                     nlmanager.nlpacket.RTMGRP_IPV6_IFADDR

        # Override the nlmanager's mac_int_to_str function to print the MACs
        # like xx:xx:xx:xx:xx:xx instead of xxxx.xxxx.xxxx
        nlmanager.nlpacket.mac_int_to_str = self.mac_int_to_str

        # Instantiate a netlink manager with a listener.
        self.nlm = ClagNlmWListener(groups, self.Log, self.Intf, self.Parser)
        self.nlm.FdbSync = self.FdbSync
        self.nlm.NeighSync = self.NeighSync

        # Enable any message debugging here.
        #self.nlm.debug_neighbor(True)
        #self.nlm.debug_link(True)

        # Add NLMSG_DONE support
        self.nlm.listener.supported_messages_add(nlmanager.nlpacket.NLMSG_DONE)

        # Prime the pump by getting all interfaces and MACs
        self.nlm.workq.put((ClagNl._GET_ALL_LINKS, None))

        # This is the main thread loop.
        while self.keep_going:

            try:
                (event, options) = self.nlm.workq.get(True, 60*60*24) # If timeout is None, signals are blocked
            except Queue.Empty:
                continue

            if event == ClagNl._GET_ALL_ADDRESSES:
                self.nlm.get_all_addresses()
            elif event == ClagNl._GET_ALL_LINKS:
                self.nlm.get_all_links()
                self.nlm.get_all_br_links()
            elif event == ClagNl._GET_ALL_NEIGHBORS:
                self.nlm.get_all_neighbors()
            elif event == ClagNl._GET_ALL_MACS:
                self.nlm.get_all_macs()
            elif event == ClagNl._GET_ALL_ROUTES:
                self.nlm.get_all_routes()
            elif event == ClagNl._SERVICE_NETLINK_QUEUE:
                self.nlm.service_netlinkq()
            elif event == ClagNl._SHUTDOWN:
                self.log_debug("NetlinkThread: shutting things down")
            elif event == ClagNl._SERVICE_ERROR:
                if 'OVERFLOW' in options:
                    self.Log.log_error("NetlinkThread: Netlink overflow")
                    self.service_overflow()
            elif event == ClagNl._GET_ALL:
                self.nlm.resync['links'] = self.nlm.get_all_links()
                self.nlm.resync['br_links'] = self.nlm.get_all_br_links()
                self.nlm.resync['macs'] = self.nlm.get_all_macs()
                self.nlm.resync['neighs'] = self.nlm.get_all_neighbors()
            else:
                raise Exception("Unsupported workq event %s" % event)

        self.log_debug("Finished execution of the thread NetlinkThread")


class ClagNlmWListener(NetlinkManagerWithListener):
    """
    The main job of this class is to override the functions of the
    NetlinkManagerWithListener class to do the things clag needs when items are
    placed on the netlink queue.
    """
    def __init__(self, groups, Log, Intf, Parser):
        use_color = Log.getHandlerStr() != "syslog"
	super(ClagNlmWListener, self).__init__(groups, use_color=use_color, error_notification=True, rcvbuf_sz=180000000)
        self.Log = Log
        self.log = Log.logger
        self.Intf = Intf
        self.Parser = Parser
        self.FdbSync = None
        self.NeighSync = None
        self.resync = {}
        self.resync_next = False

    def log_debug(self, logStr):
        self.Log.log_debug(self.Log._LOG_DEBUG_NETLINK, logStr)

    def rx_rtm_newlink(self, msg):
        self.Intf.AddInterface(self, msg)
        #self.log.debug("RXed RTM_NEWLINK seq %d, pid %d, %d bytes, for %s, state %s" %
        #               (msg.seq, msg.pid, msg.length, msg.get_attribute_value(msg.IFLA_IFNAME), "up" if msg.is_up() else "down"))

    def rx_rtm_dellink(self, msg):
        self.Intf.DelInterface(self, msg.ifindex, msg.family)
        #self.log.debug("RXed RTM_DELLINK seq %d, pid %d, %d bytes, for %s, state %s" %
        #               (msg.seq, msg.pid, msg.length, msg.get_attribute_value(msg.IFLA_IFNAME), "up" if msg.is_up() else "down"))

    def rx_rtm_newaddr(self, msg):
        if self.NeighSync and self.Parser.args.permanentMacSync:
            opcode = self.NeighSync._WORK_QUEUE_ADDR_ADD
            k_addr = msg.get_attribute_value(msg.IFA_ADDRESS)
            self.log.debug("Adding address %s/%d on %d" %
                           (k_addr, msg.prefixlen, msg.ifindex))
            self.NeighSync.fromKernWorkQueue.put((opcode, (str(k_addr), msg.family, None, msg.ifindex, None, None)))

    def rx_rtm_deladdr(self, msg):
        if self.NeighSync and self.Parser.args.permanentMacSync:
            opcode = self.NeighSync._WORK_QUEUE_ADDR_DEL
            k_addr = msg.get_attribute_value(msg.IFA_ADDRESS)
            self.log.debug("Deleting address %s/%d on %d" %
                           (k_addr, msg.prefixlen, msg.ifindex))
            self.NeighSync.fromKernWorkQueue.put((opcode, (str(k_addr), msg.family, None, msg.ifindex, None, None)))

    def rx_rtm_newneigh(self, msg):
        """
        For new dynamic MAC addresses, enqueue them for processing by the FDB sync code.
        """
        if self.FdbSync is not None and msg.family == socket.AF_BRIDGE:
            k_mac  = msg.get_attribute_value(msg.NDA_LLADDR)
            k_vlan = msg.get_attribute_value(msg.NDA_VLAN)
            k_bif  = self.Intf.GetIfName(msg.ifindex)
            k_vni  = msg.get_attribute_value(msg.NDA_VNI)
            k_dst  = msg.get_attribute_value(msg.NDA_DST)
            k_mstr = self.Intf.GetIfName(msg.get_attribute_value(msg.NDA_MASTER))
            if msg.state & msg.NUD_REACHABLE:
                self.FdbSync.fdbWorkQueue.put((self.FdbSync._WORK_QUEUE_ADD, (k_mac, k_vlan, k_bif, k_vni, k_dst, k_mstr)))
                self.log_debug("Adding MAC address %s vlan %s bif %s vni %s dst %s master %s" %
                               (k_mac, str(k_vlan), str(k_bif), str(k_vni), str(k_dst), str(k_mstr)))
            elif msg.state & msg.NUD_NOARP and self.Parser.args.permanentMacSync:
                self.FdbSync.fdbWorkQueue.put((self.FdbSync._WORK_QUEUE_STATIC_ADD, (k_mac, k_vlan, k_bif, k_vni, k_dst, k_mstr)))
                self.log_debug("Adding static MAC address %s vlan %s bif %s vni %s dst %s master %s" %
                               (k_mac, str(k_vlan), str(k_bif), str(k_vni), str(k_dst), str(k_mstr)))

        elif msg.family == socket.AF_INET or msg.family == socket.AF_INET6:
            k_mac  = msg.get_attribute_value(msg.NDA_LLADDR)
            k_dst  = msg.get_attribute_value(msg.NDA_DST)
            if self.NeighSync is not None:
                self.NeighSync.fromKernWorkQueue.put((self.NeighSync._WORK_QUEUE_ADD, (str(k_dst), msg.family, k_mac, msg.ifindex, msg.state, msg.flags)))
                self.log_debug("Adding neighbor %s idx %s mac %s" %
                               (str(k_dst), str(msg.ifindex), k_mac))

    def rx_rtm_delneigh(self, msg):
        """
        For deleted dynamic MAC addresses, enqueue them for processing by the FDB sync code.
        """
        if self.FdbSync is not None and msg.family == socket.AF_BRIDGE:
            k_mac  = msg.get_attribute_value(msg.NDA_LLADDR)
            k_vlan = msg.get_attribute_value(msg.NDA_VLAN)
            k_bif  = self.Intf.GetIfName(msg.ifindex)
            k_vni  = msg.get_attribute_value(msg.NDA_VNI)
            k_dst  = msg.get_attribute_value(msg.NDA_DST)
            k_mstr = self.Intf.GetIfName(msg.get_attribute_value(msg.NDA_MASTER))
            if msg.state & (msg.NUD_STALE | msg.NUD_REACHABLE):
                self.FdbSync.fdbWorkQueue.put((self.FdbSync._WORK_QUEUE_DELETE, (k_mac, k_vlan, k_bif, k_vni, k_dst, k_mstr)))
                self.log_debug("Deleting MAC address %s vlan %s bif %s vni %s dst %s master %s" %
                               (k_mac, str(k_vlan), str(k_bif), str(k_vni), str(k_dst), str(k_mstr)))
            elif msg.state & msg.NUD_NOARP and self.Parser.args.permanentMacSync:
                self.FdbSync.fdbWorkQueue.put((self.FdbSync._WORK_QUEUE_STATIC_DELETE, (k_mac, k_vlan, k_bif, k_vni, k_dst, k_mstr)))
                self.log_debug("Deleting static MAC address %s vlan %s bif %s vni %s dst %s master %s" %
                               (k_mac, str(k_vlan), str(k_bif), str(k_vni), str(k_dst), str(k_mstr)))

        elif msg.family == socket.AF_INET or msg.family == socket.AF_INET6:
            k_mac  = msg.get_attribute_value(msg.NDA_LLADDR)
            k_dst  = msg.get_attribute_value(msg.NDA_DST)
            if self.NeighSync is not None:
                self.NeighSync.fromKernWorkQueue.put((self.NeighSync._WORK_QUEUE_DELETE, (str(k_dst), msg.family, k_mac, msg.ifindex, msg.state, msg.flags)))
                self.log_debug("Deleting neighbor %s idx %s mac %s" %
                               (str(k_dst), str(msg.ifindex), k_mac))

    def rx_nlmsg_done(self, msg):
        if 'links' in self.resync and self.resync['links'] == msg.seq:
            self.Intf.ifInfoLock.acquire()
            idx_del_list = []
            for idx, ifDict in self.Intf.ifInfoByIndex.iteritems():
                if 'stale' in ifDict and ifDict['stale'] is True:
                    idx_del_list.append(idx)
            self.Intf.ifInfoLock.release()
            for idx in idx_del_list:
                self.Intf.DelInterface(self, idx, socket.AF_UNSPEC)
            del self.resync['links']
            self.Log.log_warn('NetlinkThread: link resync with kernel complete')
        elif 'br_links' in self.resync and self.resync['br_links'] == msg.seq:
            del self.resync['br_links']
        elif 'macs' in self.resync and self.resync['macs'] == msg.seq:
            # TODO: resync macs in 4.0 (do not wait for periodic sync)
            del self.resync['macs']
        elif 'neighs' in self.resync and self.resync['neighs'] == msg.seq:
            if self.NeighSync:
                self.NeighSync.fromKernWorkQueue.put((self.NeighSync._WORK_QUEUE_DUMP_DONE, None))
            del self.resync['neighs']

    def service_netlinkq(self):
        msg_count = {}
        processed = 0

        for (msgtype, length, flags, seq, pid, data) in self.netlinkq:
            processed += 1

            # If this is a reply to a TX message that debugs were enabled for then debug the reply
            if (seq, pid) in self.debug_seq_pid:
                debug = True
            else:
                debug = self.debug_this_packet(msgtype)

            if msgtype == nlmanager.nlpacket.RTM_NEWLINK or msgtype == nlmanager.nlpacket.RTM_DELLINK:
                msg = nlmanager.nlpacket.Link(msgtype, debug, self.log, self.Log.getHandlerStr() != "syslog")

            elif msgtype == nlmanager.nlpacket.RTM_NEWADDR or msgtype == nlmanager.nlpacket.RTM_DELADDR:
                msg = nlmanager.nlpacket.Address(msgtype, debug, self.log, self.Log.getHandlerStr() != "syslog")

            elif msgtype == nlmanager.nlpacket.RTM_NEWNEIGH or msgtype == nlmanager.nlpacket.RTM_DELNEIGH:
                msg = nlmanager.nlpacket.Neighbor(msgtype, debug, self.log, self.Log.getHandlerStr() != "syslog")

            elif msgtype == nlmanager.nlpacket.RTM_NEWROUTE or msgtype == nlmanager.nlpacket.RTM_DELROUTE:
                msg = nlmanager.nlpacket.Route(msgtype, debug, self.log, self.Log.getHandlerStr() != "syslog")

            elif msgtype == nlmanager.nlpacket.NLMSG_DONE:
                msg = nlmanager.nlpacket.Done(msgtype, debug, self.log, self.Log.getHandlerStr() != "syslog")


            else:
                self.log.warning('RXed unknown netlink message type %s' % msgtype)
                continue

            msg.decode_packet(length, flags, seq, pid, data)

            if not self.filter_permit(msg):
                continue

            if debug:
                msg.dump()

            # Only used for printing debugs about how many we RXed of each type
            if msg.msgtype not in msg_count:
                msg_count[msg.msgtype] = 0
            msg_count[msg.msgtype] += 1

            # Call the appropriate handler method based on the msgtype.  The handler
            # functions are defined in our child class.
            if msg.msgtype == nlmanager.nlpacket.RTM_NEWLINK:
                self.rx_rtm_newlink(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_DELLINK:
                self.rx_rtm_dellink(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_NEWADDR:
                self.rx_rtm_newaddr(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_DELADDR:
                self.rx_rtm_deladdr(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_NEWNEIGH:
                self.rx_rtm_newneigh(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_DELNEIGH:
                self.rx_rtm_delneigh(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_NEWROUTE:
                self.rx_rtm_newroute(msg)

            elif msg.msgtype == nlmanager.nlpacket.RTM_DELROUTE:
                self.rx_rtm_delroute(msg)

            elif msg.msgtype == nlmanager.nlpacket.NLMSG_DONE:
                self.rx_nlmsg_done(msg)

            else:
                self.log.warning('RXed unknown netlink message type %s' % msgtype)

        if processed:
            self.netlinkq = self.netlinkq[processed:]

    def get_all_addresses(self):
        family = socket.AF_UNSPEC
        debug = nlmanager.nlpacket.RTM_GETADDR in self.debug
        sequence = self.sequence.next()

        addr = nlmanager.nlpacket.Address(nlmanager.nlpacket.RTM_GETADDR, debug, self.log, self.Log.getHandlerStr() != "syslog")
        addr.flags =  nlmanager.nlpacket.NLM_F_REQUEST |  nlmanager.nlpacket.NLM_F_DUMP
        addr.body = struct.pack('Bxxxi', family, 0)
        addr.build_message(sequence, self.pid)

        if debug:
            self.debug_seq_pid[(addr.seq, addr.pid)] = True

        self.tx_nlpacket_get_response(addr)
        return sequence

    def get_all_macs(self):
        family = socket.AF_BRIDGE
        debug = nlmanager.nlpacket.RTM_GETNEIGH in self.debug
        sequence = self.sequence.next()

        neighbor = nlmanager.nlpacket.Neighbor(nlmanager.nlpacket.RTM_GETNEIGH, debug, self.log, self.Log.getHandlerStr() != "syslog")
        neighbor.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlmanager.nlpacket.NLM_F_DUMP
        neighbor.body = struct.pack('Bxxxii', family, 0, 0)
        neighbor.build_message(sequence, self.pid)

        if debug:
            self.debug_seq_pid[(neighbor.seq, neighbor.pid)] = True

        self.tx_nlpacket_get_response(neighbor)
        return sequence

    def get_all_neighbors(self):
        family = socket.AF_UNSPEC
        debug = nlmanager.nlpacket.RTM_GETNEIGH in self.debug
        sequence = self.sequence.next()

        neighbor = nlmanager.nlpacket.Neighbor(nlmanager.nlpacket.RTM_GETNEIGH, debug, self.log, self.Log.getHandlerStr() != "syslog")
        neighbor.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlmanager.nlpacket.NLM_F_DUMP
        neighbor.body = struct.pack('Bxxxii', family, 0, 0)
        neighbor.build_message(sequence, self.pid)

        if debug:
            self.debug_seq_pid[(neighbor.seq, neighbor.pid)] = True

        self.tx_nlpacket_get_response(neighbor)
        return sequence

    def get_all_links(self):
        family = socket.AF_UNSPEC
        debug = nlmanager.nlpacket.RTM_GETLINK in self.debug
        sequence = self.sequence.next()

        link = nlmanager.nlpacket.Link(nlmanager.nlpacket.RTM_GETLINK, debug, self.log, self.Log.getHandlerStr() != "syslog")
        link.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlmanager.nlpacket.NLM_F_DUMP
        link.body = struct.pack('Bxxxiii', family, 0, 0, 0)
        link.build_message(sequence, self.pid)

        if debug:
            self.debug_seq_pid[(link.seq, link.pid)] = True

        self.tx_nlpacket_get_response(link)
        return sequence

    def get_all_br_links(self):
        family = socket.AF_BRIDGE
        debug = nlmanager.nlpacket.RTM_GETLINK in self.debug
        sequence = self.sequence.next()

        link = nlmanager.nlpacket.Link(nlmanager.nlpacket.RTM_GETLINK, debug, self.log, self.Log.getHandlerStr() != "syslog")
        link.flags = nlmanager.nlpacket.NLM_F_REQUEST | nlmanager.nlpacket.NLM_F_DUMP
        link.body = struct.pack('Bxxxiii', family, 0, 0, 0)
        link.build_message(sequence, self.pid)

        if debug:
            self.debug_seq_pid[(link.seq, link.pid)] = True

        self.tx_nlpacket_get_response(link)
        return sequence

